from http import HTTPStatus
from django.http.response import HttpResponse

from . import consts, compat, exceptions


class JsonResponse(HttpResponse):
    """HTTP Response in json format
    """
    content_type = 'application/json'

    def __init__(self,
                 data,
                 *,
                 ensure_ascii=False,
                 status_code=HTTPStatus.OK,
                 **kwargs):
        if isinstance(data, str):
            data = {'detail': data}
        if hasattr(status_code, 'value'):
            status_code = status_code.value
        kwargs['content_type'] = self.content_type
        content = compat.json.dumps(data, ensure_ascii=ensure_ascii)
        super().__init__(content=content, status=status_code, **kwargs)

    @classmethod
    def from_exception(cls,
                       exc: exceptions.APIError,
                       status_code=HTTPStatus.INTERNAL_SERVER_ERROR,
                       user_hint=''):
        """Generate response from APIError
        """
        if not user_hint:
            user_hint = getattr(exc, 'user_hint', '')
        if not user_hint:
            user_hint = consts.ERROR_MSG.get(status_code, '')
        status_code = getattr(exc, 'status_code', status_code)
        return cls(user_hint, status_code=status_code)

    @classmethod
    def created(cls, data, **kwargs):
        kwargs['status_code'] = HTTPStatus.CREATED
        return cls(data, **kwargs)

    @classmethod
    def not_found(cls, data, **kwargs):
        kwargs['status_code'] = HTTPStatus.NOT_FOUND
        return cls(data, **kwargs)


APIResponse = JsonResponse



from django.test import TestCase


class PermissionTestCase(TestCase):
    def test_basic_auth_required(self):
        response = self.client.get('/basic_auth')
        resp_data = response.json()
        self.assertEqual(response.status_code, 401)
        self.assertEqual(resp_data['detail'], 'basic_auth required')

        response = self.client.get('/basic_auth',
                                   HTTP_AUTHORIZATION='Basic: foo')
        resp_data = response.json()
        self.assertEqual(response.status_code, 200)
        self.assertEqual(resp_data['msg'], 'hello world')

    def test_owner_required(self):
        response = self.client.get('/blogs/2')
        resp_data = response.json()
        self.assertEqual(resp_data['detail'], 'not owner')

        response = self.client.get('/blogs/1')
        resp_data = response.json()
        self.assertEqual(resp_data['pk'], 1)



import time

from django.conf import settings
from django.contrib.auth import get_user_model
from django.test import TestCase

from simple_django_api.jwt.auth import generate_token

User = get_user_model()


class JwtTestCase(TestCase):
    @classmethod
    def setUpTestData(cls):
        # Set up data for the whole TestCase
        cls.user = User.objects.create_user(
            'john',
            'lennon@thebeatles.com',
            'johnpassword',
        )

    def test_no_auth(self):
        # no auth
        response = self.client.get('/auth')
        self.assertEqual(response.status_code, 401)

    def test_non_bearer_auth(self):
        # non bearer auth
        response = self.client.get('/auth', HTTP_AUTHORIZATION='adasdadsa')
        resp_data = response.json()
        self.assertEqual(response.status_code, 401)
        self.assertEqual(resp_data['detail'], 'NO_TOKEN')

    def test_invalid_token(self):
        # invalid token
        response = self.client.get('/auth',
                                   HTTP_AUTHORIZATION='Bearer adasdadsa')
        resp_data = response.json()
        self.assertEqual(response.status_code, 401)
        self.assertEqual(resp_data['detail'], 'DECODE_ERROR')

    def test_valid_token(self):
        # valid token
        token = generate_token(self.__class__.user)
        response = self.client.get('/auth',
                                   HTTP_AUTHORIZATION=f'Bearer {token}')
        resp_data = response.json()
        self.assertEqual(response.status_code, 200)
        self.assertEqual(resp_data['username'], 'john')

    def test_broken_token(self):
        token = generate_token(self.__class__.user)
        new_token = token + 'some_data'
        response = self.client.get('/auth',
                                   HTTP_AUTHORIZATION=f'Bearer {new_token}')
        resp_data = response.json()
        self.assertEqual(response.status_code, 401)
        self.assertEqual(resp_data['detail'], 'DECODE_ERROR')

    def test_expired_token(self):
        token = generate_token(self.__class__.user)
        time.sleep(60 * settings.API_JWT_EXPIRATION_MINUTES + 1)
        response = self.client.get('/auth',
                                   HTTP_AUTHORIZATION=f'Bearer {token}')
        resp_data = response.json()
        self.assertEqual(response.status_code, 401)
        self.assertEqual(resp_data['detail'], 'EXPIRED')



import django
from django.conf import settings
from django.core import management

from .demo_app import settings as project_settings


def setup_django():
    settings.configure(
        **{
            key: getattr(project_settings, key)
            for key in dir(project_settings) if key.isupper()
        })
    django.setup()


collect_ignore = []
if django.__version__[:3] > '2.2':
    collect_ignore.append('request/django_2x/')
else:
    collect_ignore.append('request/django_3x/')

setup_django()
management.call_command('makemigrations')
management.call_command('migrate')



from simple_django_api.views import APIView
from .permissions import (BasicAuthRequired, login_required, owner_required)


class BasicAuthView(APIView):
    method_perms = {'GET': BasicAuthRequired}

    def get(self, request):
        return {'msg': 'hello world'}


class AuthView(APIView):
    method_perms = {'get': login_required}

    def get(self, request):
        return {'username': request.user.username}


class BlogDetailView(APIView):
    method_perms = {'get': owner_required}

    def get(self, request, pk=None):
        return {'pk': pk}



from io import BytesIO
from types import MethodType

from django.core.handlers.wsgi import WSGIRequest as OriginWSGIRequest
from django.http.request import HttpRequest, QueryDict
from django.http.multipartparser import MultiPartParserError
from django.utils.datastructures import MultiValueDict

from . import compat, exceptions


def parse_json_body(input_data, *, encoding=None):
    encoding = encoding or 'utf8'
    try:
        if hasattr(input_data, 'decode'):
            input_data = input_data.decode(encoding)
        return compat.json.loads(input_data)
    except (compat.json.JSONDecodeError, ValueError) as exc:
        raise exceptions.InvalidRequestBody() from exc


# this code is part of django.http.request.HttpRequest
def _load_post_and_files(self):
    """Populate self._post and self._files
    if the content-type is a form type"""
    if self.method not in ['POST', 'PUT', 'PATCH']:
        self._post, self._files = QueryDict(
            encoding=self._encoding), MultiValueDict()
        return
    if self._read_started and not hasattr(self, '_body'):
        self._mark_post_parse_error()
        return

    if self.content_type == 'multipart/form-data':
        if hasattr(self, '_body'):
            # Use already read data
            data = BytesIO(self._body)
        else:
            data = self
        try:
            self._post, self._files = self.parse_file_upload(self.META, data)
        except MultiPartParserError:
            # An error occurred while parsing POST data. Since when
            # formatting the error the request handler might access
            # self.POST, set self._post and self._file to prevent
            # attempts to parse POST data again.
            self._mark_post_parse_error()
            raise
    elif self.content_type == 'application/x-www-form-urlencoded':
        self._post, self._files = QueryDict(
            self.body, encoding=self._encoding), MultiValueDict()
    elif self.content_type == 'application/json':
        self._post = parse_json_body(self.body, encoding=self._encoding)
        self._files = MultiValueDict()
    else:
        self._post, self._files = QueryDict(
            encoding=self._encoding), MultiValueDict()


class Request:
    def __init__(self, raw_request=None):
        if raw_request is None:
            raw_request = HttpRequest()
        raw_request._load_post_and_files = MethodType(_load_post_and_files,
                                                      raw_request)
        self._raw_request = raw_request

    @property
    def data(self):
        return self._raw_request.POST

    @property
    def query_params(self):
        return self._raw_request.GET

    @property
    def files(self):
        return self._raw_request.files

    def __getattr__(self, name):
        return getattr(self._raw_request, name)

    def __setattr__(self, name, value):
        key_lst = {
            '_raw_request',
        }
        if name in key_lst:
            super().__setattr__(name, value)
        else:
            setattr(self._raw_request, name, value)

    def __repr__(self):
        return repr(self._raw_request)

    def __iter__(self):
        return iter(self._raw_request)


class WSGIRequest(Request):
    def __init__(self, environ, raw_request=None):
        if not raw_request:
            raw_request = OriginWSGIRequest(environ)
        super().__init__(raw_request=raw_request)

    def __repr__(self):
        return repr(self._raw_request)



import enum


class TokenCase(enum.Enum):
    OK = enum.auto()
    NO_TOKEN = enum.auto()
    EXPIRED = enum.auto()
    DECODE_ERROR = enum.auto()
    INVALID_TOKEN = enum.auto()
    MISSING_KEY = enum.auto()



from simple_django_api import permissions
from simple_django_api.response import APIResponse
from simple_django_api.jwt.consts import TokenCase
from .exceptions import Unauthorized


class BasicAuthRequired(permissions.BasePermission):
    def __call__(self, request):
        if not request.headers.get('Authorization'):
            raise Unauthorized(user_hint='basic_auth required')


def login_required(view_cls):
    def inner(request):
        if not request.user.is_authenticated:
            raise Unauthorized(user_hint=request.jwt_info['case'].name)

    return inner


def owner_required(view_cls):
    def inner(request, pk=None, **kwargs):
        if pk != 1:
            return APIResponse('not owner')

    return inner



"""
You can writer permission in two ways:

.. code:: python

    class LoginRequired:
        def __init__(self, view_cls):
            self.view_cls = view_cls

        def __call__(self, request, **kwargs):
            if not request.user.is_authenticated:
                raise exceptions.Unauthorized()


.. code:: python

    def login_required(view):
        def inner(request, **kwargs):
            if not request.user.is_authenticated:
                raise exceptions.Unauthorized()

        return inner
"""
import abc
from . import exceptions


class BasePermission(abc.ABC):
    def __init__(self, view_cls):
        self.view = self.view_cls = view_cls

    @abc.abstractmethod
    def __call__(self, request, **kwargs):
        return True


class LoginRequired(BasePermission):
    def __call__(self, request, **kwargs):
        if not request.user.is_authenticated:
            raise exceptions.Unauthorized()


class SuperUserRequired(BasePermission):
    def __call__(self, request, **kwargs):
        if not request.user.is_superuser:
            raise exceptions.Forbidden()


def login_required(view_cls):
    def inner(request, **kwargs):
        if not request.user.is_authenticated:
            raise exceptions.Unauthorized()

    return inner



from django.conf import settings


class FallbackSettings:
    def __init__(self, **default):
        self._default = {key.upper(): value for key, value in default.items()}

    def __getattr__(self, name):
        try:
            return getattr(settings, name)
        except AttributeError as exc:
            if name in self._default:
                return self._default[name]
            raise AttributeError from exc



from http import HTTPStatus
import logging

from django.conf import settings
from django.http import HttpResponse
from django.views import View
from django.views.decorators.csrf import csrf_exempt
from django.utils.decorators import method_decorator

try:
    from webargs import ValidationError
except ImportError:
    from .exceptions import ValidationError

from .request import Request
from .response import APIResponse
from .utils.logging import LoggingContext


class APIView(View):
    method_perms = None
    logger = logging.getLogger(__name__)

    def _check_permission(self, request, **kwargs):
        default_perms = getattr(settings, 'API_DEFAULT_PERMS', [])
        method_perms = self.method_perms or {}
        perm_lst = method_perms.get(request.method.upper(), None)
        if perm_lst is None:
            perm_lst = method_perms.get(request.method.lower(), None)
        if perm_lst is None:
            perm_lst = default_perms
        if not isinstance(perm_lst, (list, tuple)):
            perm_lst = [perm_lst]
        for perm in perm_lst:
            perm_checker = perm(self)
            resp = perm_checker(request, **kwargs)
            if resp:
                return resp

    def _dispatch(self, request, *args, **kwargs):
        return super().dispatch(request, *args, **kwargs)

    @method_decorator(csrf_exempt)
    def dispatch(self, request, *args, **kwargs):
        self.request = request = Request(raw_request=request)
        try:
            resp = self._check_permission(request, **kwargs)
            if not resp:
                resp = self._dispatch(request, *args, **kwargs)
            if not isinstance(resp, HttpResponse):
                resp = APIResponse(resp)
            return resp
        except Exception as exc:
            exc_handler = getattr(settings, 'API_VIEW_EXCEPTION_HANDLER',
                                  self.exception_handler)
            return exc_handler(request, exc)

    def exception_handler(self, request, exc):
        options = {}
        if isinstance(exc, ValidationError):
            options['status_code'] = HTTPStatus.BAD_REQUEST
            options['user_hint'] = exc.messages
        self.log_exception(request, exc)
        return APIResponse.from_exception(exc, **options)

    def log_exception(self, request, exc):
        context = LoggingContext()
        context['method'] = request.method
        context['path'] = request.path
        context['query_params'] = request.query_params
        context['data'] = request.data
        context['user'] = str(request.user)
        context['exc'] = exc
        level = getattr(exc, 'logging_level', logging.ERROR)
        exc_info = level >= logging.ERROR
        self.logger.log(level, context, exc_info=exc_info)



"""settings for demo app
"""
ALLOWED_HOSTS = '*'
DEBUG_PROPAGATE_EXCEPTIONS = True
DATABASES = {
    'default': {
        'ENGINE': 'django.db.backends.sqlite3',
        'NAME': ':memory:'
    }
}
SITE_ID = 1
SECRET_KEY = 'not very secret in tests'
STATIC_URL = '/static/'
ROOT_URLCONF = 'tests.demo_app.urls'
TEMPLATE_LOADERS = (
    'django.template.loaders.filesystem.Loader',
    'django.template.loaders.app_directories.Loader',
)
MIDDLEWARE = (
    'django.middleware.security.SecurityMiddleware',
    'django.contrib.sessions.middleware.SessionMiddleware',
    'django.middleware.common.CommonMiddleware',
    'django.middleware.csrf.CsrfViewMiddleware',
    'django.contrib.auth.middleware.AuthenticationMiddleware',
    'django.contrib.messages.middleware.MessageMiddleware',
    'django.middleware.clickjacking.XFrameOptionsMiddleware',
    'simple_django_api.jwt.middleware.AuthenticationMiddleware',
)
INSTALLED_APPS = (
    'django.contrib.auth',
    'django.contrib.contenttypes',
    'django.contrib.sessions',
    'django.contrib.sites',
    'django.contrib.messages',
    'django.contrib.staticfiles',
)
PASSWORD_HASHERS = ('django.contrib.auth.hashers.MD5PasswordHasher', )
LOGGING = {
    'version': 1,
    'disable_existing_loggers': True,
    'handlers': {
        'console': {
            'class': 'logging.StreamHandler',
        },
    },
    'root': {
        'handlers': ['console'],
        'level': 'WARNING',
    },
    'loggers': {
        'app': {
            'handlers': ['console'],
            'level': 'INFO',
            'propagate': True,
        },
    },
}
API_JWT_SECRET_KEY = 'some key'
API_JWT_EXPIRATION_MINUTES = 0.05



from __future__ import unicode_literals

from io import BytesIO

from django.http import (RawPostDataException, UnreadablePostError)
from django.test import SimpleTestCase
from django.test.client import FakePayload
from django.utils.http import urlencode

from simple_django_api.request import WSGIRequest

HTTP_METHODS_WITH_BODY = ('POST', 'PUT', 'PATCH')


class OriginRequestTest(SimpleTestCase):
    def test_read_after_value(self):
        """
        Reading from request is allowed after accessing request contents as
        POST or body.
        """
        for method in HTTP_METHODS_WITH_BODY:
            payload = FakePayload('name=value')
            request = WSGIRequest({
                'REQUEST_METHOD': method,
                'CONTENT_TYPE': 'application/x-www-form-urlencoded',
                'CONTENT_LENGTH': len(payload),
                'wsgi.input': payload
            })
            self.assertEqual(request.POST, {'name': ['value']})
            self.assertEqual(request.body, b'name=value')
            self.assertEqual(request.read(), b'name=value')

    def test_value_after_read(self):
        """
        Construction of POST or body is not allowed after reading
        from request.
        """
        for method in HTTP_METHODS_WITH_BODY:
            payload = FakePayload('name=value')
            request = WSGIRequest({
                'REQUEST_METHOD': method,
                'CONTENT_TYPE': 'application/x-www-form-urlencoded',
                'CONTENT_LENGTH': len(payload),
                'wsgi.input': payload
            })
            self.assertEqual(request.read(2), b'na')
            with self.assertRaises(RawPostDataException):
                request.body
            self.assertEqual(request.POST, {})

    def test_non_ascii_POST(self):
        for method in HTTP_METHODS_WITH_BODY:
            payload = FakePayload(urlencode({'key': 'España'}))
            request = WSGIRequest({
                'REQUEST_METHOD': method,
                'CONTENT_LENGTH': len(payload),
                'CONTENT_TYPE': 'application/x-www-form-urlencoded',
                'wsgi.input': payload,
            })
            self.assertEqual(request.POST, {'key': ['España']})

    def test_alternate_charset_POST(self):
        """
        Test a POST with non-utf-8 payload encoding.
        """
        for method in HTTP_METHODS_WITH_BODY:
            payload = FakePayload(
                urlencode({
                    'key': 'España'.encode('latin-1')
                }))
            request = WSGIRequest({
                'REQUEST_METHOD': method,
                'CONTENT_LENGTH': len(payload),
                'CONTENT_TYPE': 'application/x-www-form-urlencoded; charset=iso-8859-1',
                'wsgi.input': payload,
            })
            self.assertEqual(request.POST, {'key': ['España']})

    def test_body_after_POST_multipart_form_data(self):
        """
        Reading body after parsing multipart/form-data is not allowed
        """
        # Because multipart is used for large amounts of data i.e. file uploads,
        # we don't want the data held in memory twice, and we don't want to
        # silence the error by setting body = '' either.
        for method in HTTP_METHODS_WITH_BODY:
            payload = FakePayload("\r\n".join([
                '--boundary', 'Content-Disposition: form-data; name="name"',
                '', 'value', '--boundary--'
                ''
            ]))
            request = WSGIRequest({
                'REQUEST_METHOD': method,
                'CONTENT_TYPE': 'multipart/form-data; boundary=boundary',
                'CONTENT_LENGTH': len(payload),
                'wsgi.input': payload
            })
            self.assertEqual(request.POST, {'name': ['value']})
            with self.assertRaises(RawPostDataException):
                request.body

    def test_body_after_POST_multipart_related(self):
        """
        Reading body after parsing multipart that isn't form-data is allowed
        """
        # Ticket #9054
        # There are cases in which the multipart data is related instead of
        # being a binary upload, in which case it should still be accessible
        # via body.
        for method in HTTP_METHODS_WITH_BODY:
            payload_data = b"\r\n".join([
                b'--boundary', b'Content-ID: id; name="name"', b'', b'value',
                b'--boundary--'
                b''
            ])
            payload = FakePayload(payload_data)
            request = WSGIRequest({
                'REQUEST_METHOD': method,
                'CONTENT_TYPE': 'multipart/related; boundary=boundary',
                'CONTENT_LENGTH': len(payload),
                'wsgi.input': payload
            })
            self.assertEqual(request.POST, {})
            self.assertEqual(request.body, payload_data)

    def test_POST_multipart_with_content_length_zero(self):
        """
        Multipart POST requests with Content-Length >= 0 are valid and need to be handled.
        """
        # According to:
        # http://www.w3.org/Protocols/rfc2616/rfc2616-sec14.html#sec14.13
        # Every request.POST with Content-Length >= 0 is a valid request,
        # this test ensures that we handle Content-Length == 0.
        for method in HTTP_METHODS_WITH_BODY:
            payload = FakePayload("\r\n".join([
                '--boundary', 'Content-Disposition: form-data; name="name"',
                '', 'value', '--boundary--'
                ''
            ]))
            request = WSGIRequest({
                'REQUEST_METHOD': method,
                'CONTENT_TYPE': 'multipart/form-data; boundary=boundary',
                'CONTENT_LENGTH': 0,
                'wsgi.input': payload
            })
            self.assertEqual(request.POST, {})

    def test_POST_binary_only(self):
        for method in HTTP_METHODS_WITH_BODY:
            payload = b'\r\n\x01\x00\x00\x00ab\x00\x00\xcd\xcc,@'
            environ = {
                'REQUEST_METHOD': method,
                'CONTENT_TYPE': 'application/octet-stream',
                'CONTENT_LENGTH': len(payload),
                'wsgi.input': BytesIO(payload)
            }
            request = WSGIRequest(environ)
            self.assertEqual(request.POST, {})
            self.assertEqual(request.FILES, {})
            self.assertEqual(request.body, payload)

            # Same test without specifying content-type
            environ.update({
                'CONTENT_TYPE': '',
                'wsgi.input': BytesIO(payload)
            })
            request = WSGIRequest(environ)
            self.assertEqual(request.POST, {})
            self.assertEqual(request.FILES, {})
            self.assertEqual(request.body, payload)

    def test_read_by_lines(self):
        for method in HTTP_METHODS_WITH_BODY:
            payload = FakePayload('name=value')
            request = WSGIRequest({
                'REQUEST_METHOD': 'POST',
                'CONTENT_TYPE': 'application/x-www-form-urlencoded',
                'CONTENT_LENGTH': len(payload),
                'wsgi.input': payload
            })
            self.assertEqual(list(request), [b'name=value'])

    def test_POST_after_body_read(self):
        """
        POST should be populated even if body is read first
        """
        for method in HTTP_METHODS_WITH_BODY:
            payload = FakePayload('name=value')
            request = WSGIRequest({
                'REQUEST_METHOD': 'POST',
                'CONTENT_TYPE': 'application/x-www-form-urlencoded',
                'CONTENT_LENGTH': len(payload),
                'wsgi.input': payload
            })
            request.body    # evaluate
            self.assertEqual(request.POST, {'name': ['value']})

    def test_POST_after_body_read_and_stream_read(self):
        """
        POST should be populated even if body is read first, and then
        the stream is read second.
        """
        for method in HTTP_METHODS_WITH_BODY:
            payload = FakePayload('name=value')
            request = WSGIRequest({
                'REQUEST_METHOD': method,
                'CONTENT_TYPE': 'application/x-www-form-urlencoded',
                'CONTENT_LENGTH': len(payload),
                'wsgi.input': payload
            })
            request.body    # evaluate
            self.assertEqual(request.read(1), b'n')
            self.assertEqual(request.POST, {'name': ['value']})

    def test_POST_after_body_read_and_stream_read_multipart(self):
        """
        POST should be populated even if body is read first, and then
        the stream is read second. Using multipart/form-data instead of urlencoded.
        """
        for method in HTTP_METHODS_WITH_BODY:
            payload = FakePayload("\r\n".join([
                '--boundary', 'Content-Disposition: form-data; name="name"',
                '', 'value', '--boundary--'
                ''
            ]))
            request = WSGIRequest({
                'REQUEST_METHOD': method,
                'CONTENT_TYPE': 'multipart/form-data; boundary=boundary',
                'CONTENT_LENGTH': len(payload),
                'wsgi.input': payload
            })
            request.body    # evaluate
            # Consume enough data to mess up the parsing:
            self.assertEqual(request.read(13), b'--boundary\r\nC')
            self.assertEqual(request.POST, {'name': ['value']})

    def test_POST_immutable_for_mutipart(self):
        """
        MultiPartParser.parse() leaves request.POST immutable.
        """
        for method in HTTP_METHODS_WITH_BODY:
            payload = FakePayload("\r\n".join([
                '--boundary',
                'Content-Disposition: form-data; name="name"',
                '',
                'value',
                '--boundary--',
            ]))
            request = WSGIRequest({
                'REQUEST_METHOD': method,
                'CONTENT_TYPE': 'multipart/form-data; boundary=boundary',
                'CONTENT_LENGTH': len(payload),
                'wsgi.input': payload,
            })
            self.assertFalse(request.POST._mutable)

    def test_POST_connection_error(self):
        """
        If wsgi.input.read() raises an exception while trying to read() the
        POST, the exception should be identifiable (not a generic IOError).
        """

        class ExplodingBytesIO(BytesIO):
            def read(self, len=0):
                raise IOError("kaboom!")

        for method in HTTP_METHODS_WITH_BODY:
            payload = b'name=value'
            request = WSGIRequest({
                'REQUEST_METHOD': method,
                'CONTENT_TYPE': 'application/x-www-form-urlencoded',
                'CONTENT_LENGTH': len(payload),
                'wsgi.input': ExplodingBytesIO(payload)
            })

            with self.assertRaises(UnreadablePostError):
                request.body

    def test_set_encoding_clears_POST(self):
        for method in HTTP_METHODS_WITH_BODY:
            payload = FakePayload('name=Hello Günter')
            request = WSGIRequest({
                'REQUEST_METHOD': method,
                'CONTENT_TYPE': 'application/x-www-form-urlencoded',
                'CONTENT_LENGTH': len(payload),
                'wsgi.input': payload,
            })
            self.assertEqual(request.POST, {'name': ['Hello Günter']})
            request.encoding = 'iso-8859-16'
            self.assertEqual(request.POST, {'name': ['Hello GĂŒnter']})

    def test_FILES_connection_error(self):
        """
        If wsgi.input.read() raises an exception while trying to read() the
        FILES, the exception should be identifiable (not a generic IOError).
        """

        class ExplodingBytesIO(BytesIO):
            def read(self, len=0):
                raise IOError("kaboom!")

        for method in HTTP_METHODS_WITH_BODY:
            payload = b'x'
            request = WSGIRequest({
                'REQUEST_METHOD': method,
                'CONTENT_TYPE': 'multipart/form-data; boundary=foo_',
                'CONTENT_LENGTH': len(payload),
                'wsgi.input': ExplodingBytesIO(payload)
            })

            with self.assertRaises(UnreadablePostError):
                request.FILES


class PatchedRequestsTests(SimpleTestCase):
    def test_read_after_value(self):
        """
        Reading from request is allowed after accessing request contents as
        data or body.
        """
        for method in HTTP_METHODS_WITH_BODY:
            payload = FakePayload('{"name": "value"}')
            request = WSGIRequest({
                'REQUEST_METHOD': method,
                'CONTENT_TYPE': 'application/json',
                'CONTENT_LENGTH': len(payload),
                'wsgi.input': payload
            })
            self.assertEqual(request.data, {'name': 'value'})
            self.assertEqual(request.body, b'{"name": "value"}')
            self.assertEqual(request.read(), b'{"name": "value"}')

    def test_value_after_read(self):
        """
        Construction of POST or body is not allowed after reading
        from request.
        """
        for method in HTTP_METHODS_WITH_BODY:
            payload = FakePayload('{"name": "value"}')
            request = WSGIRequest({
                'REQUEST_METHOD': 'POST',
                'CONTENT_TYPE': 'application/json',
                'CONTENT_LENGTH': len(payload),
                'wsgi.input': payload
            })
            self.assertEqual(request.read(2), b'{"')
            with self.assertRaises(RawPostDataException):
                request.body
            self.assertEqual(request.data, {})

    def test_non_ascii_POST(self):
        for method in HTTP_METHODS_WITH_BODY:
            payload = FakePayload('{"name": "佚名"}')
            request = WSGIRequest({
                'REQUEST_METHOD': method,
                'CONTENT_LENGTH': len(payload),
                'CONTENT_TYPE': 'application/json',
                'wsgi.input': payload,
            })
            self.assertEqual(request.data, {'name': '佚名'})

    def test_alternate_charset_POST(self):
        """
        Test a POST with non-utf-8 payload encoding.
        """
        for method in HTTP_METHODS_WITH_BODY:
            payload = FakePayload('{"name": "佚名"}'.encode('GBK'))
            request = WSGIRequest({
                'REQUEST_METHOD': method,
                'CONTENT_LENGTH': len(payload),
                'CONTENT_TYPE': 'application/json; charset=GBK',
                'wsgi.input': payload,
            })
            self.assertEqual(request.data, {'name': '佚名'})

    def test_POST_after_body_read(self):
        """
        POST should be populated even if body is read first
        """
        for method in HTTP_METHODS_WITH_BODY:
            payload = FakePayload('{"name": "佚名"}')
            request = WSGIRequest({
                'REQUEST_METHOD': method,
                'CONTENT_TYPE': 'application/json',
                'CONTENT_LENGTH': len(payload),
                'wsgi.input': payload
            })
            request.body    # evaluate
            self.assertEqual(request.data, {'name': '佚名'})

    def test_POST_after_body_read_and_stream_read(self):
        """
        POST should be populated even if body is read first, and then
        the stream is read second.
        """
        for method in HTTP_METHODS_WITH_BODY:
            payload = FakePayload('{"name": "佚名"}')
            request = WSGIRequest({
                'REQUEST_METHOD': 'POST',
                'CONTENT_TYPE': 'application/json',
                'CONTENT_LENGTH': len(payload),
                'wsgi.input': payload
            })
            request.body    # evaluate
            self.assertEqual(request.read(1), b'{')
            self.assertEqual(request.POST, {'name': '佚名'})
            self.assertEqual(request.data, {'name': '佚名'})

    def test_POST_connection_error(self):
        """
        If wsgi.input.read() raises an exception while trying to read() the
        POST, the exception should be identifiable (not a generic IOError).
        """

        class ExplodingBytesIO(BytesIO):
            def read(self, len=0):
                raise IOError("kaboom!")

        for method in HTTP_METHODS_WITH_BODY:
            payload = '{"name": "佚名"}'.encode('utf8')
            request = WSGIRequest({
                'REQUEST_METHOD': method,
                'CONTENT_TYPE': 'application/json',
                'CONTENT_LENGTH': len(payload),
                'wsgi.input': ExplodingBytesIO(payload)
            })

        with self.assertRaises(UnreadablePostError):
            request.body

    def test_set_encoding_clears_POST(self):
        for method in HTTP_METHODS_WITH_BODY:
            payload = FakePayload('{"name": "佚名"}')
            request = WSGIRequest({
                'REQUEST_METHOD': 'POST',
                'CONTENT_TYPE': 'application/json',
                'CONTENT_LENGTH': len(payload),
                'wsgi.input': payload,
            })
            self.assertEqual(request.data, {'name': '佚名'})
            request.encoding = 'GBK'
            self.assertEqual(request.POST, {'name': '浣氬悕'})




from http import HTTPStatus

ERROR_MSG = {
    HTTPStatus.NOT_FOUND: 'not found',
    HTTPStatus.BAD_REQUEST: 'bad request',
    HTTPStatus.FORBIDDEN: 'permission denied',
    HTTPStatus.INTERNAL_SERVER_ERROR: 'internal server error',
}



import logging
from http import HTTPStatus
from . import consts


class APIError(Exception):
    status_code = HTTPStatus.INTERNAL_SERVER_ERROR
    logging_level = logging.ERROR

    def __init__(self,
                 *,
                 user_hint='',
                 logging_hint='',
                 status_code=None,
                 logging_level=None):
        cls = self.__class__
        self.user_hint = user_hint or consts.ERROR_MSG.get(status_code, '')
        self.logging_hint = logging_hint or self.user_hint
        self.status_code = status_code or cls.status_code
        self.logging_level = logging_level or cls.logging_level

    def __str__(self):
        cls_name = self.__class__.__name__
        return f'{cls_name}: {self.user_hint}'

    def __repr__(self):
        cls_name = self.__class__.__name__
        return f'{cls_name}: {self.logging_hint}'


class InternalError(APIError):
    status_code = HTTPStatus.INTERNAL_SERVER_ERROR
    logging_level = logging.ERROR


class InvalidRequestBody(APIError):
    status_code = HTTPStatus.BAD_REQUEST


class ParamsError(APIError):
    status_code = HTTPStatus.BAD_REQUEST
    logging_level = logging.WARNING


class ValidationError(ParamsError):
    messages = consts.ERROR_MSG.get(HTTPStatus.BAD_REQUEST)


class Unauthorized(APIError):
    status_code = HTTPStatus.UNAUTHORIZED
    logging_level = logging.WARNING


class Forbidden(APIError):
    status_code = HTTPStatus.FORBIDDEN
    logging_level = logging.WARNING


class NotFound(APIError):
    status_code = HTTPStatus.NOT_FOUND
    logging_level = logging.WARNING



from datetime import datetime, timedelta
from dataclasses import dataclass

from django.contrib.auth import get_user_model
from django.utils.functional import SimpleLazyObject
import jwt

from .settings import settings
from .consts import TokenCase
from ..utils.string import ensure_str


def get_user(user_pk=None):
    """
    Return the user model instance associated with the given pk.
    If no user is retrieved, return an instance of `AnonymousUser`.
    """
    from django.contrib.auth.models import AnonymousUser
    user = None
    if user_pk:
        user = get_user_model().objects.filter(pk=user_pk).first()
    return user or AnonymousUser()


@dataclass
class JwtConfig:
    key: str
    algorithm: str
    verify_exp: bool
    auth_cookie: str
    expiration_minutes: (float, int)
    user_pk_key: str


def get_jwt_config():
    return JwtConfig(
        key=settings.API_JWT_SECRET_KEY,
        algorithm=settings.API_JWT_ALGORITHM,
        verify_exp=settings.API_JWT_VERIFY_EXPIRATION,
        auth_cookie=settings.API_JWT_AUTH_COOKIE,
        expiration_minutes=settings.API_JWT_EXPIRATION_MINUTES,
        user_pk_key='sub',
    )


JWT_CONFIG = SimpleLazyObject(get_jwt_config)


def encode(payload: dict) -> str:
    """Accept payload and generate token string
    """
    token = jwt.encode(payload, JWT_CONFIG.key, JWT_CONFIG.algorithm)
    return ensure_str(token)


def decode(token: str) -> dict:
    """receive token and get payload
    """
    options = {'verify_exp': JWT_CONFIG.verify_exp}
    return jwt.decode(
        token,
        JWT_CONFIG.key,
        options=options,
        algorithms=[JWT_CONFIG.algorithm],
    )


def get_token_from_request(request):
    """get token string from request
    token may stored in cookie or header `AUTHORIZATION`
    """
    if JWT_CONFIG.auth_cookie:
        authorization = request.COOKIES.get(JWT_CONFIG.auth_cookie, b'')
    else:
        authorization = request.META.get('HTTP_AUTHORIZATION', b'')
    if hasattr(authorization, 'decode'):
        try:
            authorization = authorization.decode()
        except UnicodeDecodeError:
            return None, 'decode error'
    auth = authorization.split()
    if len(auth) == 2 and auth[0].lower() == 'bearer':
        return auth[1], ''
    return None, 'bad format'


def generate_token(user):
    """generate token for user
    """
    delta = timedelta(minutes=JWT_CONFIG.expiration_minutes)
    expire_at = datetime.utcnow() + delta
    payload = {'exp': expire_at, JWT_CONFIG.user_pk_key: user.pk}
    return encode(payload)


def authenticate(request):
    """Return the user model instance associated with the given request
    If no user is retrieved, return an instance of `AnonymousUser`
    """
    token, _ = get_token_from_request(request)
    jwt_info = {
        'token': token,
        'case': TokenCase.OK,
        'payload': None,
    }
    if not token:
        jwt_info['case'] = TokenCase.NO_TOKEN
        return get_user(), jwt_info
    try:
        payload = decode(token)
        user_pk = payload[JWT_CONFIG.user_pk_key]
        return get_user(user_pk=user_pk), jwt_info
    except jwt.ExpiredSignatureError:
        jwt_info['case'] = TokenCase.EXPIRED
    except jwt.DecodeError:
        jwt_info['case'] = TokenCase.DECODE_ERROR
    except jwt.InvalidTokenError:
        jwt_info['case'] = TokenCase.INVALID_TOKEN
    except KeyError:
        jwt_info['case'] = TokenCase.MISSING_KEY
    return get_user(), jwt_info



from simple_django_api.utils.settings import FallbackSettings

settings = FallbackSettings(
    API_JWT_ALGORITHM='HS256',
    API_JWT_AUTH_COOKIE='',
    API_JWT_VERIFY_EXPIRATION=True,
    API_JWT_EXPIRATION_MINUTES=10,
)



from django.urls import path
from . import views

urlpatterns = [
    path('basic_auth', views.BasicAuthView.as_view()),
    path('auth', views.AuthView.as_view()),
    path('blogs/<int:pk>', views.BlogDetailView.as_view()),
]



from collections import OrderedDict


class LoggingContext(OrderedDict):
    def __init__(self, delimiter=' '):
        self.delimiter = delimiter

    def __str__(self):
        parts = [f'[{key}: {value}]' for key, value in self.items()]
        return self.delimiter.join(parts)



from django.utils.functional import SimpleLazyObject
from . import auth


def authenticate(request):
    user, jwt_info = auth.authenticate(request)
    request._cached_user = user
    request._jwt_info = jwt_info


def get_user(request):
    # other middleware may already set request._user
    if not hasattr(request, '_jwt_info'):
        authenticate(request)
    return request._cached_user


def get_jwt_info(request):
    if not hasattr(request, '_jwt_info'):
        authenticate(request)
    return request._jwt_info


class AuthenticationMiddleware:
    def __init__(self, get_response):
        self.get_response = get_response

    def __call__(self, request):
        if not hasattr(request, 'jwt_info'):
            request.user = SimpleLazyObject(lambda: get_user(request))
            request.jwt_info = SimpleLazyObject(lambda: get_jwt_info(request))
        response = self.get_response(request)
        return response
